
import Foundation

// MARK: - Expression

@resultBuilder
public struct ValidationRulesBuilder {
    
    public static func buildExpression<T : MaybeAllowValidationRule>(_ expression: T,
                                                                     file: StaticString = #file,
                                                                     function: StaticString = #function,
                                                                     line: Int = #line) -> PartialValidationRules<T.Input> {

        return PartialValidationRules(validator: ValidationRuleValidator(rule: expression, file: file, function: function, line: line))
    }

    public static func buildExpression<T : MaybeDenyValidationRule>(_ expression: T,
                                                                    file: StaticString = #file,
                                                                    function: StaticString = #function,
                                                                    line: Int = #line) -> PartialValidationRules<T.Input> {

        return PartialValidationRules(validator: ValidationRuleValidator(rule: expression, file: file, function: function, line: line))
    }

    public static func buildExpression<T : FinalValidationRule>(_ expression: T,
                                                                file: StaticString = #file,
                                                                function: StaticString = #function,
                                                                line: Int = #line) -> ValidationRules<T.Input> {

        return ValidationRules(validator: ValidationRuleValidator(rule: expression, file: file, function: function, line: line))
    }

    public static func buildExpression<T : PartialValidator>(_ expression: T) -> PartialValidationRules<T.Input> {
        return expression.rules
    }

    public static func buildExpression<T : Validator>(_ expression: T) -> ValidationRules<T.Input> {
        return expression.rules
    }

    public static func buildExpression<Input>(_ expression: PartialValidationRules<Input>) -> PartialValidationRules<Input> {
        return expression
    }

    public static func buildExpression<Input>(_ expression: ValidationRules<Input>) -> ValidationRules<Input> {
        return expression
    }

}

// MARK: - Either

extension ValidationRulesBuilder {

    public static func buildEither<Input>(first component: PartialValidationRules<Input>) -> PartialValidationRules<Input> {
        return component
    }

    public static func buildEither<Input>(second component: PartialValidationRules<Input>) -> PartialValidationRules<Input> {
        return component
    }

    public static func buildEither<Input>(first component: ValidationRules<Input>) -> ValidationRules<Input> {
        return component
    }

    public static func buildEither<Input>(second component: ValidationRules<Input>) -> ValidationRules<Input> {
        return component
    }

}

// MARK: - Optionals

extension ValidationRulesBuilder {

    public static func buildOptional<Input>(_ component: PartialValidationRules<Input>?) -> PartialValidationRules<Input> {
        return component ?? PartialValidationRules(validator: InternalValidator())
    }

    public static func buildOptional<Input>(_ component: ValidationRules<Input>?) -> PartialValidationRules<Input> {
        return PartialValidationRules(validator: component?.validator ?? InternalValidator())
    }

}

// MARK: - Array

extension ValidationRulesBuilder {

    public static func buildArray<Input>(_ components: [PartialValidationRules<Input>]) -> PartialValidationRules<Input> {
        return PartialValidationRules(validator: CompoundValidator(validators: components.map(\.validator)))
    }

}

// MARK: - Bloks

extension ValidationRulesBuilder {

    public static func buildBlock<Input>() -> PartialValidationRules<Input> {
        return PartialValidationRules(validator: InternalValidator())
    }

    public static func buildBlock<Input>(_ components: PartialValidationRules<Input>...) -> PartialValidationRules<Input> {
        return PartialValidationRules(validator: CompoundValidator(validators: components.map(\.validator)))
    }

    public static func buildBlock<Input>(_ component: ValidationRules<Input>) -> ValidationRules<Input> {
        return component
    }

    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: PartialValidationRules<Input>, _ c88: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator).followed(by: c88.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: PartialValidationRules<Input>, _ c88: PartialValidationRules<Input>, _ c89: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator).followed(by: c88.validator).followed(by: c89.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: PartialValidationRules<Input>, _ c88: PartialValidationRules<Input>, _ c89: PartialValidationRules<Input>, _ c90: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator).followed(by: c88.validator).followed(by: c89.validator).followed(by: c90.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: PartialValidationRules<Input>, _ c88: PartialValidationRules<Input>, _ c89: PartialValidationRules<Input>, _ c90: PartialValidationRules<Input>, _ c91: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator).followed(by: c88.validator).followed(by: c89.validator).followed(by: c90.validator).followed(by: c91.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: PartialValidationRules<Input>, _ c88: PartialValidationRules<Input>, _ c89: PartialValidationRules<Input>, _ c90: PartialValidationRules<Input>, _ c91: PartialValidationRules<Input>, _ c92: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator).followed(by: c88.validator).followed(by: c89.validator).followed(by: c90.validator).followed(by: c91.validator).followed(by: c92.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: PartialValidationRules<Input>, _ c88: PartialValidationRules<Input>, _ c89: PartialValidationRules<Input>, _ c90: PartialValidationRules<Input>, _ c91: PartialValidationRules<Input>, _ c92: PartialValidationRules<Input>, _ c93: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator).followed(by: c88.validator).followed(by: c89.validator).followed(by: c90.validator).followed(by: c91.validator).followed(by: c92.validator).followed(by: c93.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: PartialValidationRules<Input>, _ c88: PartialValidationRules<Input>, _ c89: PartialValidationRules<Input>, _ c90: PartialValidationRules<Input>, _ c91: PartialValidationRules<Input>, _ c92: PartialValidationRules<Input>, _ c93: PartialValidationRules<Input>, _ c94: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator).followed(by: c88.validator).followed(by: c89.validator).followed(by: c90.validator).followed(by: c91.validator).followed(by: c92.validator).followed(by: c93.validator).followed(by: c94.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: PartialValidationRules<Input>, _ c88: PartialValidationRules<Input>, _ c89: PartialValidationRules<Input>, _ c90: PartialValidationRules<Input>, _ c91: PartialValidationRules<Input>, _ c92: PartialValidationRules<Input>, _ c93: PartialValidationRules<Input>, _ c94: PartialValidationRules<Input>, _ c95: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator).followed(by: c88.validator).followed(by: c89.validator).followed(by: c90.validator).followed(by: c91.validator).followed(by: c92.validator).followed(by: c93.validator).followed(by: c94.validator).followed(by: c95.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: PartialValidationRules<Input>, _ c88: PartialValidationRules<Input>, _ c89: PartialValidationRules<Input>, _ c90: PartialValidationRules<Input>, _ c91: PartialValidationRules<Input>, _ c92: PartialValidationRules<Input>, _ c93: PartialValidationRules<Input>, _ c94: PartialValidationRules<Input>, _ c95: PartialValidationRules<Input>, _ c96: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator).followed(by: c88.validator).followed(by: c89.validator).followed(by: c90.validator).followed(by: c91.validator).followed(by: c92.validator).followed(by: c93.validator).followed(by: c94.validator).followed(by: c95.validator).followed(by: c96.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: PartialValidationRules<Input>, _ c88: PartialValidationRules<Input>, _ c89: PartialValidationRules<Input>, _ c90: PartialValidationRules<Input>, _ c91: PartialValidationRules<Input>, _ c92: PartialValidationRules<Input>, _ c93: PartialValidationRules<Input>, _ c94: PartialValidationRules<Input>, _ c95: PartialValidationRules<Input>, _ c96: PartialValidationRules<Input>, _ c97: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator).followed(by: c88.validator).followed(by: c89.validator).followed(by: c90.validator).followed(by: c91.validator).followed(by: c92.validator).followed(by: c93.validator).followed(by: c94.validator).followed(by: c95.validator).followed(by: c96.validator).followed(by: c97.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: PartialValidationRules<Input>, _ c88: PartialValidationRules<Input>, _ c89: PartialValidationRules<Input>, _ c90: PartialValidationRules<Input>, _ c91: PartialValidationRules<Input>, _ c92: PartialValidationRules<Input>, _ c93: PartialValidationRules<Input>, _ c94: PartialValidationRules<Input>, _ c95: PartialValidationRules<Input>, _ c96: PartialValidationRules<Input>, _ c97: PartialValidationRules<Input>, _ c98: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator).followed(by: c88.validator).followed(by: c89.validator).followed(by: c90.validator).followed(by: c91.validator).followed(by: c92.validator).followed(by: c93.validator).followed(by: c94.validator).followed(by: c95.validator).followed(by: c96.validator).followed(by: c97.validator).followed(by: c98.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: PartialValidationRules<Input>, _ c88: PartialValidationRules<Input>, _ c89: PartialValidationRules<Input>, _ c90: PartialValidationRules<Input>, _ c91: PartialValidationRules<Input>, _ c92: PartialValidationRules<Input>, _ c93: PartialValidationRules<Input>, _ c94: PartialValidationRules<Input>, _ c95: PartialValidationRules<Input>, _ c96: PartialValidationRules<Input>, _ c97: PartialValidationRules<Input>, _ c98: PartialValidationRules<Input>, _ c99: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator).followed(by: c88.validator).followed(by: c89.validator).followed(by: c90.validator).followed(by: c91.validator).followed(by: c92.validator).followed(by: c93.validator).followed(by: c94.validator).followed(by: c95.validator).followed(by: c96.validator).followed(by: c97.validator).followed(by: c98.validator).followed(by: c99.validator))
    }
    
    public static func buildBlock<Input>(_ c0: PartialValidationRules<Input>, _ c1: PartialValidationRules<Input>, _ c2: PartialValidationRules<Input>, _ c3: PartialValidationRules<Input>, _ c4: PartialValidationRules<Input>, _ c5: PartialValidationRules<Input>, _ c6: PartialValidationRules<Input>, _ c7: PartialValidationRules<Input>, _ c8: PartialValidationRules<Input>, _ c9: PartialValidationRules<Input>, _ c10: PartialValidationRules<Input>, _ c11: PartialValidationRules<Input>, _ c12: PartialValidationRules<Input>, _ c13: PartialValidationRules<Input>, _ c14: PartialValidationRules<Input>, _ c15: PartialValidationRules<Input>, _ c16: PartialValidationRules<Input>, _ c17: PartialValidationRules<Input>, _ c18: PartialValidationRules<Input>, _ c19: PartialValidationRules<Input>, _ c20: PartialValidationRules<Input>, _ c21: PartialValidationRules<Input>, _ c22: PartialValidationRules<Input>, _ c23: PartialValidationRules<Input>, _ c24: PartialValidationRules<Input>, _ c25: PartialValidationRules<Input>, _ c26: PartialValidationRules<Input>, _ c27: PartialValidationRules<Input>, _ c28: PartialValidationRules<Input>, _ c29: PartialValidationRules<Input>, _ c30: PartialValidationRules<Input>, _ c31: PartialValidationRules<Input>, _ c32: PartialValidationRules<Input>, _ c33: PartialValidationRules<Input>, _ c34: PartialValidationRules<Input>, _ c35: PartialValidationRules<Input>, _ c36: PartialValidationRules<Input>, _ c37: PartialValidationRules<Input>, _ c38: PartialValidationRules<Input>, _ c39: PartialValidationRules<Input>, _ c40: PartialValidationRules<Input>, _ c41: PartialValidationRules<Input>, _ c42: PartialValidationRules<Input>, _ c43: PartialValidationRules<Input>, _ c44: PartialValidationRules<Input>, _ c45: PartialValidationRules<Input>, _ c46: PartialValidationRules<Input>, _ c47: PartialValidationRules<Input>, _ c48: PartialValidationRules<Input>, _ c49: PartialValidationRules<Input>, _ c50: PartialValidationRules<Input>, _ c51: PartialValidationRules<Input>, _ c52: PartialValidationRules<Input>, _ c53: PartialValidationRules<Input>, _ c54: PartialValidationRules<Input>, _ c55: PartialValidationRules<Input>, _ c56: PartialValidationRules<Input>, _ c57: PartialValidationRules<Input>, _ c58: PartialValidationRules<Input>, _ c59: PartialValidationRules<Input>, _ c60: PartialValidationRules<Input>, _ c61: PartialValidationRules<Input>, _ c62: PartialValidationRules<Input>, _ c63: PartialValidationRules<Input>, _ c64: PartialValidationRules<Input>, _ c65: PartialValidationRules<Input>, _ c66: PartialValidationRules<Input>, _ c67: PartialValidationRules<Input>, _ c68: PartialValidationRules<Input>, _ c69: PartialValidationRules<Input>, _ c70: PartialValidationRules<Input>, _ c71: PartialValidationRules<Input>, _ c72: PartialValidationRules<Input>, _ c73: PartialValidationRules<Input>, _ c74: PartialValidationRules<Input>, _ c75: PartialValidationRules<Input>, _ c76: PartialValidationRules<Input>, _ c77: PartialValidationRules<Input>, _ c78: PartialValidationRules<Input>, _ c79: PartialValidationRules<Input>, _ c80: PartialValidationRules<Input>, _ c81: PartialValidationRules<Input>, _ c82: PartialValidationRules<Input>, _ c83: PartialValidationRules<Input>, _ c84: PartialValidationRules<Input>, _ c85: PartialValidationRules<Input>, _ c86: PartialValidationRules<Input>, _ c87: PartialValidationRules<Input>, _ c88: PartialValidationRules<Input>, _ c89: PartialValidationRules<Input>, _ c90: PartialValidationRules<Input>, _ c91: PartialValidationRules<Input>, _ c92: PartialValidationRules<Input>, _ c93: PartialValidationRules<Input>, _ c94: PartialValidationRules<Input>, _ c95: PartialValidationRules<Input>, _ c96: PartialValidationRules<Input>, _ c97: PartialValidationRules<Input>, _ c98: PartialValidationRules<Input>, _ c99: PartialValidationRules<Input>, _ c100: ValidationRules<Input>) -> ValidationRules<Input> {
        return ValidationRules(validator: c0.validator.followed(by: c1.validator).followed(by: c2.validator).followed(by: c3.validator).followed(by: c4.validator).followed(by: c5.validator).followed(by: c6.validator).followed(by: c7.validator).followed(by: c8.validator).followed(by: c9.validator).followed(by: c10.validator).followed(by: c11.validator).followed(by: c12.validator).followed(by: c13.validator).followed(by: c14.validator).followed(by: c15.validator).followed(by: c16.validator).followed(by: c17.validator).followed(by: c18.validator).followed(by: c19.validator).followed(by: c20.validator).followed(by: c21.validator).followed(by: c22.validator).followed(by: c23.validator).followed(by: c24.validator).followed(by: c25.validator).followed(by: c26.validator).followed(by: c27.validator).followed(by: c28.validator).followed(by: c29.validator).followed(by: c30.validator).followed(by: c31.validator).followed(by: c32.validator).followed(by: c33.validator).followed(by: c34.validator).followed(by: c35.validator).followed(by: c36.validator).followed(by: c37.validator).followed(by: c38.validator).followed(by: c39.validator).followed(by: c40.validator).followed(by: c41.validator).followed(by: c42.validator).followed(by: c43.validator).followed(by: c44.validator).followed(by: c45.validator).followed(by: c46.validator).followed(by: c47.validator).followed(by: c48.validator).followed(by: c49.validator).followed(by: c50.validator).followed(by: c51.validator).followed(by: c52.validator).followed(by: c53.validator).followed(by: c54.validator).followed(by: c55.validator).followed(by: c56.validator).followed(by: c57.validator).followed(by: c58.validator).followed(by: c59.validator).followed(by: c60.validator).followed(by: c61.validator).followed(by: c62.validator).followed(by: c63.validator).followed(by: c64.validator).followed(by: c65.validator).followed(by: c66.validator).followed(by: c67.validator).followed(by: c68.validator).followed(by: c69.validator).followed(by: c70.validator).followed(by: c71.validator).followed(by: c72.validator).followed(by: c73.validator).followed(by: c74.validator).followed(by: c75.validator).followed(by: c76.validator).followed(by: c77.validator).followed(by: c78.validator).followed(by: c79.validator).followed(by: c80.validator).followed(by: c81.validator).followed(by: c82.validator).followed(by: c83.validator).followed(by: c84.validator).followed(by: c85.validator).followed(by: c86.validator).followed(by: c87.validator).followed(by: c88.validator).followed(by: c89.validator).followed(by: c90.validator).followed(by: c91.validator).followed(by: c92.validator).followed(by: c93.validator).followed(by: c94.validator).followed(by: c95.validator).followed(by: c96.validator).followed(by: c97.validator).followed(by: c98.validator).followed(by: c99.validator).followed(by: c100.validator))
    }
    
}
